from .parsers.c_parser import extract_comments as iter_comments_c
from .parsers.common import Comment, Span, FileError, UnterminatedCommentError
from .parsers.css_parser import extract_comments as iter_comments_css
from .parsers.go_parser import extract_comments as iter_comments_go
from .parsers.html_parser import extract_comments as iter_comments_html
from .parsers.js_parser import extract_comments as iter_comments_js
from .parsers.python_parser import extract_comments as iter_comments_python
from .parsers.ruby_parser import extract_comments as iter_comments_ruby
from .parsers.shell_parser import extract_comments as iter_comments_shell



# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-License-Identifier: MIT

import re

from commie.parsers.common import Comment, Span


def matchGroupToComment(match:re.Match, groupName:str, multiline:bool) -> Comment:

  fullSpan = match.span()

  fullText = match.group(0)
  innerText = match.group(groupName)
  textStart = fullText.index(innerText)
  assert textStart>=0

  return Comment(
    text_span=Span(fullSpan[0]+textStart, fullSpan[0]+textStart+len(innerText)),
    markup_span=Span(fullSpan[0], fullSpan[1]),
    multiline=multiline)


# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

from typing import Iterable

from commie.parsers import common
from commie.parsers.common import Comment, Span


def extract_comments(goCode: str) -> Iterable[Comment]:
  """Extracts a list of comments from the given Go source code.

  Comments are represented with the Comment class found in the common module.
  Go comments come in two forms, single and multi-line comments.
    - Single-line comments begin with '//' and continue to the end of line.
    - Multi-line comments begin with '/*' and end with '*/' and can span
      multiple lines of code. If a multi-line comment does not terminate
      before EOF is reached, then an exception is raised.
  Go comments are not allowed to start in a string or rune literal. This
  module makes sure to watch out for those.

  https://golang.org/ref/spec#Comments

  Args:
    goCode: String containing code to extract comments from.
  Returns:
    Python list of common.Comment in the order that they appear in the code.
  Raises:
    common.UnterminatedCommentError: Encountered an unterminated multi-line
      comment.
  """

  DEFAULT = 0
  AFTER_SLASH = 1
  IN_SINGLE_LINE_COMMENT = 2
  IN_MULTI_LINE_COMMENT = 3
  IN_MULTI_LINE_COMMENT_ASTERISK = 4
  IN_STRING = 5
  IN_STRING_ESCAPING = 6

  state = DEFAULT
  current_comment = ''

  line_counter = 1
  string_char = ''

  position = -1
  comment_start_pos = None

  for position, char in enumerate(goCode):
    if state == DEFAULT:
      # Waiting for comment start character or beginning of
      # string or rune literal.
      if char == '/':
        state = AFTER_SLASH
      elif char in ('"', "'", '`'):
        string_char = char
        state = IN_STRING
    elif state == AFTER_SLASH:
      # Found comment start character, classify next character and
      # determine if single or multi-line comment.
      if char == '/':
        state = IN_SINGLE_LINE_COMMENT
        comment_start_pos = position-1
      elif char == '*':
        comment_start = line_counter
        state = IN_MULTI_LINE_COMMENT
        comment_start_pos = position-1
      else:
        state = DEFAULT
    elif state == IN_SINGLE_LINE_COMMENT:
      # In single-line comment, read characters util EOL.
      if char == '\n':
        yield Comment(Span(comment_start_pos, position+1), Span(comment_start_pos+2, position+1), False)
        #yield Comment(current_comment, comment_start_pos, position+1, False)
        current_comment = ''
        state = 0
      else:
        current_comment += char
    elif state == IN_MULTI_LINE_COMMENT:
      # In multi-line comment, add characters until '*' is
      # encountered.
      if char == '*':
        state = IN_MULTI_LINE_COMMENT_ASTERISK
      else:
        current_comment += char
    elif state == IN_MULTI_LINE_COMMENT_ASTERISK:
      # In multi-line comment with asterisk found. Determine if
      # comment is ending.
      if char == '/':
        yield Comment(Span(comment_start_pos, position+1), Span(comment_start_pos+2, position-1), True)
        current_comment = ''
        state = DEFAULT
      else:
        current_comment += '*'
        # Care for multiple '*' in a row
        if char != '*':
          current_comment += char
          state = IN_MULTI_LINE_COMMENT
    elif state == IN_STRING:
      # In string literal, expect literal end or escape character.
      if char == string_char:
        state = DEFAULT
      elif char == '\\':
        state = IN_STRING_ESCAPING
    elif state == IN_STRING_ESCAPING:
      # In string literal, escaping current char.
      state = IN_STRING
    if char == '\n':
      line_counter += 1

  # EOF

  if state in (IN_MULTI_LINE_COMMENT, IN_MULTI_LINE_COMMENT_ASTERISK):
    raise common.UnterminatedCommentError()
  if state == IN_SINGLE_LINE_COMMENT:
    # was in single-line comment
    yield Comment(Span(comment_start_pos, position+1), Span(comment_start_pos+2, position+1), False)

    #yield Comment(current_comment, comment_start_pos, position+1, False)



# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

import io
import tokenize
from typing import NamedTuple, Iterable

from commie.parsers.common import Comment, Span


class PosToken(NamedTuple):
  tokenType: int
  text: str
  start: int
  end: int


def postokenize(infile: io.BytesIO) -> Iterable[PosToken]:
  # based on https://stackoverflow.com/a/62761208 (CC BY-SA 4.0)

  # Used to track starting position of each line.
  # Note that tokenize starts line numbers at 1 and column numbers at 0
  offsets = [0]

  def wrapped_readline():
    # Function used to wrap calls to infile.readline(); stores current
    # stream position at the beginning of each line.
    offsets.append(infile.tell())
    return infile.readline()

  # For each returned token, substitute type with exact_type and
  # add token boundaries as stream positions
  for t in tokenize.tokenize(wrapped_readline):
    startline, startcol = t.start
    endline, endcol = t.end
    yield PosToken(t.exact_type, t.string,
                   offsets[startline] + startcol,
                   offsets[endline] + endcol)


def extract_comments(code: str) -> Iterable[Comment]:

  """Extracts a list of comments from the given Python script.
  Comments are identified using the tokenize module. Does not include function,
  class, or module docstrings. All comments are single line comments.
  Args:
    code: String containing code to extract comments from.
  Returns:
    Python list of common.Comment in the order that they appear in the code.
  Raises:
    tokenize.TokenError
  """

  for token in postokenize(io.BytesIO(code.encode())):
    if token.tokenType == tokenize.COMMENT:
      yield Comment(
        text_span=Span(token.start+1, token.end),
        markup_span=Span(token.start, token.end),
        multiline=False
      )


if __name__ == "__main__":

  def experiment():
    from pathlib import Path
    extract_comments(Path(__file__).read_text())


  experiment()



# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

import unittest
from typing import List

from .. import iter_comments_js
from ..parsers.common import Comment, UnterminatedCommentError


def commentsToList(code:str) -> List[Comment]:
  return list(iter_comments_js(code))


class JsParserTest(unittest.TestCase):

  def testSingleLineComment(self):
    code = '// single line comment'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "// single line comment")
    self.assertEqual(comments[0].text_span.extract(code), " single line comment")
    self.assertEqual(comments[0].multiline, False)

  def testLineCommentInSingleQuotedLiteral(self):
    code = "msg = '// this is not a comment'"
    comments = commentsToList(code)
    self.assertEqual(len(comments), 0)

  def testLineCommentInDoubleQuotedLiteral(self):
    code = 'msg = "// this is not a comment"'
    comments = commentsToList(code)
    self.assertEqual(len(comments), 0)

  def testMultiLineComment(self):
    code = '/* multiline\ncomment */'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)
    self.assertEqual(comments[0].markup_span.extract(code), '/* multiline\ncomment */')
    self.assertEqual(comments[0].text_span.extract(code), " multiline\ncomment ")
    self.assertEqual(comments[0].multiline, True)


  def testMultiLineCommentWithStars(self):
    code = "/***************/"
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)
    self.assertEqual(comments[0].markup_span.extract(code), '/***************/')
    self.assertEqual(comments[0].text_span.extract(code), "*************")
    self.assertEqual(comments[0].multiline, True)

  def testMultiLineCommentInSingleLiteral(self):
    code = "msg = '/* This is not a\\nmultiline comment */'"
    comments = commentsToList(code)
    self.assertEqual(len(comments), 0)

  def testMultiLineCommentInDoubleLiteral(self):
    code = 'msg = "/* This is not a\\nmultiline comment */"'
    comments = commentsToList(code)
    self.assertEqual(len(comments), 0)

  def testMultiLineCommentUnterminated(self):
    code = 'a = 1 /* Unterminated\\n comment'
    with self.assertRaises(UnterminatedCommentError):
      commentsToList(code)

  def testMultiple(self):
    code = """

    /*
     * this is a fancy comment
     * one more line of it
     */

    let forgetIt = prompt("What?");

    // bye!

    """

    comments = commentsToList(code)

    self.assertEqual(len(comments), 2)
    self.assertEqual(comments[0].multiline, True)
    self.assertEqual(comments[1].multiline, False)



# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

import re
from typing import Iterable

from commie.parsers.common import Comment
from commie.parsers.helper import matchGroupToComment


def extract_comments(rubyCode: str) -> Iterable[Comment]:
  """Extracts a list of comments from the given Ruby source code.

  Comments are represented with the Comment class found in the common module.

  Ruby comments start with a '#' character and run to the end of the line,
  http://ruby-doc.com/docs/ProgrammingRuby.

  Args:
    rubyCode: String containing code to extract comments from.
  Returns:
    Python list of common.Comment in the order that they appear in the code..
  """
  pattern = r"""
    (?P<literal> ([\"'])((?:\\\2|(?:(?!\2)).)*)(\2)) |
    (?P<single> \#(?P<single_content>.*?)$)
  """
  compiled = re.compile(pattern, re.VERBOSE | re.MULTILINE)

  comments = []
  for match in compiled.finditer(rubyCode):
    kind = match.lastgroup
    span = match.span(0)

    if kind == "single":
      yield matchGroupToComment(match, "single_content", False)

  return comments


# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-License-Identifier: MIT

import re
from typing import Iterable

from commie.parsers import common
from commie.parsers.common import Comment
from commie.parsers.helper import matchGroupToComment


def extract_comments(cssCode:str) -> Iterable[Comment]:

  pattern = r"""
    (?P<comment> /\*(?P<content>(.|\n)*?)?\*/) |
    (?P<error> /\*(.*)?)
  """

  for match in re.finditer(pattern, cssCode, flags=re.VERBOSE | re.MULTILINE):

    kind = match.lastgroup

    if kind == "comment":
      yield matchGroupToComment(match, "content", True)

    elif kind == "error":
      raise common.UnterminatedCommentError()


# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

import unittest
from typing import List

from .. import iter_comments_c
from ..parsers.common import Comment, UnterminatedCommentError


def commentsToList(code:str) -> List[Comment]:
  return list(iter_comments_c(code))


class CParserTest(unittest.TestCase):

  def testSimpleMain(self):
    code = "// this is a comment\nint main() {\nreturn 0;\n}\n"
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "// this is a comment")
    self.assertEqual(comments[0].text_span.extract(code), " this is a comment")
    self.assertEqual(comments[0].multiline, False)

  def testSingleLineComment(self):
    code = '// single line comment'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '// single line comment')
    self.assertEqual(comments[0].text_span.extract(code), " single line comment")
    self.assertEqual(comments[0].multiline, False)

  def testSingleLineCommentInStringLiteral(self):
    code = 'char* msg = "// this is not a comment"'
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testMultiLineComment(self):
    code = '/* multiline\ncomment */'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '/* multiline\ncomment */')
    self.assertEqual(comments[0].text_span.extract(code), " multiline\ncomment ")
    self.assertEqual(comments[0].multiline, True)

  def testMultiLineCommentWithStars(self):
    code = "/***************/"
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '/***************/')
    self.assertEqual(comments[0].text_span.extract(code), "*************")
    self.assertEqual(comments[0].multiline, True)

  def testMultiLineCommentInStringLiteral(self):
    code = 'char* msg = "/* This is not a\\nmultiline comment */"'
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testMultiLineCommentUnterminated(self):
    code = 'int a = 1; /* Unterminated\\n comment'
    with self.assertRaises(UnterminatedCommentError):
      commentsToList(code)

  def testMultipleMultilineComments(self):
    code = '/* abc */ /* 123 */'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 2)

    self.assertEqual(comments[0].markup_span.extract(code), '/* abc */')
    self.assertEqual(comments[0].text_span.extract(code), " abc ")
    self.assertEqual(comments[0].multiline, True)

    self.assertEqual(comments[1].markup_span.extract(code), '/* 123 */')
    self.assertEqual(comments[1].text_span.extract(code), " 123 ")
    self.assertEqual(comments[1].multiline, True)


  def testStringThenComment(self):
    code = r'"" /* "abc */'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '/* "abc */')
    self.assertEqual(comments[0].text_span.extract(code), ' "abc ')
    self.assertEqual(comments[0].multiline, True)


  def testCommentStartInsideEscapedQuotesInStringLiteral(self):
    # TODO(#27): Re-enable test.
    # code = r'" \" /* \" "'
    # comments = c_parser.extract_comments(code)
    # self.assertEqual(comments, [])
    pass


  def testStringEscapedBackslashCharacter(self):
    code = r'"\\"'
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testTwoStringsFollowedByComment(self):
    code = r'"""" // foo'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '// foo')
    self.assertEqual(comments[0].text_span.extract(code), " foo")
    self.assertEqual(comments[0].multiline, False)

  def testCommentedMultilineComment(self):
    code = '''// What if i start a /* here
    int main(){return 0;}
    // and ended it here */'''

    comments = commentsToList(code)

    self.assertEqual(len(comments), 2)

    self.assertEqual(comments[0].markup_span.extract(code), '// What if i start a /* here')
    self.assertEqual(comments[0].text_span.extract(code), " What if i start a /* here")
    self.assertEqual(comments[0].multiline, False)

    self.assertEqual(comments[1].markup_span.extract(code), '// and ended it here */')
    self.assertEqual(comments[1].text_span.extract(code), " and ended it here */")
    self.assertEqual(comments[1].multiline, False)


  def testMultilineCommentedComment(self):
    code = '''/*// here
    int main(){return 0;}
    */// and ended it here */'''
    comments = commentsToList(code)

    self.assertEqual(len(comments), 2)

    self.assertEqual(comments[0].markup_span.extract(code), "/*// here\n    int main(){return 0;}\n    */")
    self.assertEqual(comments[0].text_span.extract(code), "// here\n    int main(){return 0;}\n    ")
    self.assertEqual(comments[0].multiline, True)

    self.assertEqual(comments[1].markup_span.extract(code), '// and ended it here */')
    self.assertEqual(comments[1].text_span.extract(code), " and ended it here */")
    self.assertEqual(comments[1].multiline, False)



# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

"""This module provides methods for parsing comments from HTML family languages.

Works with:
  HTML
  XML
  SGML
"""

import re
from typing import Iterable

from commie.parsers import common
from commie.parsers.common import Comment
from commie.parsers.helper import matchGroupToComment


def extract_comments(htmlCode: str) -> Iterable[Comment]:
  """Extracts a list of comments from the given HTML family source code.

  Comments are represented with the Comment class found in the common module.
  HTML family comments come in one form, comprising all text within '<!--' and
  '-->' markers. Comments cannot be nested.

  Args:
    htmlCode: String containing code to extract comments from.
  Returns:
    Python list of common.Comment in the order that they appear in the code..
  Raises:
    common.UnterminatedCommentError: Encountered an unterminated multi-line
      comment.
  """
  pattern = r"""
    (?P<literal> (\"([^\"\n])*\")+) |
    (?P<single> <!--(?P<single_content>.*?)-->) |
    (?P<multi> <!--(?P<multi_content>(.|\n)*?)?-->) |
    (?P<error> <!--(.*)?)
  """
  compiled = re.compile(pattern, re.VERBOSE | re.MULTILINE)

  for match in compiled.finditer(htmlCode):

    kind = match.lastgroup

    if kind == "single":
      yield matchGroupToComment(match, "single_content", False)
    elif kind == "multi":
      yield matchGroupToComment(match, "multi_content", True)
    elif kind == "error":
      raise common.UnterminatedCommentError()




# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

from typing import NamedTuple


class Error(Exception):
  """Base Error class for all comment parsers."""


class FileError(Error):
  """Raised if there is an issue reading a file."""


class UnterminatedCommentError(Error):
  """Raised if an Unterminated multi-line comment is encountered."""


class Span(NamedTuple):
  start: int
  end: int

  def extract(self, text: str):
    return text[self.start:self.end]


class Comment:
  """Represents comments found in a source code string."""

  def __init__(self, markup_span: Span, text_span: Span, multiline: bool):

    self.markup_span: Span = markup_span
    self.text_span: Span = text_span

    self.multiline = multiline

  def __repr__(self):
    return f"Comment({self.markup_span}, {self.text_span}, {self.multiline})"

  def __eq__(self, other):
    if isinstance(other, self.__class__):
      if self.__dict__ == other.__dict__:
        return True
    return False



# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

import unittest
from typing import List

from .. import iter_comments_html
from ..parsers.common import Comment, UnterminatedCommentError


def commentsToList(code:str) -> List[Comment]:
  return list(iter_comments_html(code))


class HtmlParserTest(unittest.TestCase):

  def testComment(self):
    code = '<!--comment-->'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "<!--comment-->")
    self.assertEqual(comments[0].text_span.extract(code), "comment")
    self.assertEqual(comments[0].multiline, False)

  def testMultilineComment(self):
    code = '<!--multi-line\ncomment-->'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '<!--multi-line\ncomment-->')
    self.assertEqual(comments[0].text_span.extract(code), "multi-line\ncomment")
    self.assertEqual(comments[0].multiline, True)

  def testTwoSeparateSingleComment(self):
    code = '<!--comment1-->\n<!--comment2-->'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 2)

    self.assertEqual(comments[0].markup_span.extract(code), '<!--comment1-->')
    self.assertEqual(comments[0].text_span.extract(code), "comment1")
    self.assertEqual(comments[0].multiline, False)

    self.assertEqual(comments[1].markup_span.extract(code), '<!--comment2-->')
    self.assertEqual(comments[1].text_span.extract(code), "comment2")
    self.assertEqual(comments[1].multiline, False)

  def testLayeredComment(self):
    code = '<!-- comment<!-- -->'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '<!-- comment<!-- -->')
    self.assertEqual(comments[0].text_span.extract(code), " comment<!-- ")
    self.assertEqual(comments[0].multiline, False)

  def testNonGreedyComment(self):
    code = '<!--i am a comment--> not a comment -->'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '<!--i am a comment-->')
    self.assertEqual(comments[0].text_span.extract(code), "i am a comment")
    self.assertEqual(comments[0].multiline, False)

  def testSideBySideComment(self):
    code = '<!--comment1--> ... <!--comment2-->'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 2)

    self.assertEqual(comments[0].markup_span.extract(code), '<!--comment1-->')
    self.assertEqual(comments[0].text_span.extract(code), "comment1")
    self.assertEqual(comments[0].multiline, False)

    self.assertEqual(comments[1].markup_span.extract(code), '<!--comment2-->')
    self.assertEqual(comments[1].text_span.extract(code), "comment2")
    self.assertEqual(comments[1].multiline, False)

  def testUnterminatedComment(self):
    code = '<!--invalid'
    with self.assertRaises(UnterminatedCommentError):
      commentsToList(code)

  def testLonelyTerminator(self):
    code = 'not a comment-->'
    comments = commentsToList(code)
    self.assertEqual(comments, [])



# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

import unittest
from typing import List

from commie import iter_comments_css
from ..parsers.common import Comment


def commentsToList(code:str) -> List[Comment]:
  return list(iter_comments_css(code))


class CssParserTest(unittest.TestCase):

  def testSimpleMain(self):
    code = ".cssClass { /* i am\n a comment! */ }"
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "/* i am\n a comment! */")
    self.assertEqual(comments[0].text_span.extract(code), " i am\n a comment! ")
    self.assertEqual(comments[0].multiline, True)

  def testIncLeft(self):
    code = ".cssClass { /* /* i am a comment! */ }"
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "/* /* i am a comment! */")
    self.assertEqual(comments[0].text_span.extract(code), " /* i am a comment! ")
    self.assertEqual(comments[0].multiline, True)

  def testIncRight(self):
    code = ".cssClass { /* i am a comment! */ */ }"
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "/* i am a comment! */")
    self.assertEqual(comments[0].text_span.extract(code), " i am a comment! ")
    self.assertEqual(comments[0].multiline, True)

  def testThreeComments(self):
    code = """
    
    /* THIS IS A CSS CODE
       WITH MULTILINE COMMENTS */
    
    body {margin: 0; /* comment inside */}
    p {margin: 1px; ) /* comment outside */
    
    """
    comments = commentsToList(code)

    self.assertEqual(len(comments), 3)


# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

from typing import Iterable

from commie.parsers.common import Comment, Span


def extract_comments(code: str) -> Iterable[Comment]:

  """Extracts a list of comments from the given shell script.
  Comments are represented with the Comment class found in the common module.
  Shell script comments only come in one form, single-line. Single line
  comments start with an unquoted or unescaped '#' and continue on until the
  end of the line. A quoted '#' is one that is located within a pair of
  matching single or double quote marks. An escaped '#' is one that is
  immediately preceeded by a backslash '\'
  Args:
    code: String containing code to extract comments from.
  Returns:
    Python list of common.Comment in the order that they appear in the code.
  """

  DEFAULT = 0
  IN_COMMENT = 1
  IN_STRING = 2
  ESCAPING_CHAR_INSIDE_STRING = 3
  ESCAPING_CHAR_OUTSIDE_OF_STRING = 4

  state = DEFAULT
  string_char = ''
  current_comment_text = ''
  line_counter = 1

  comment_start_pos = None
  position = -1

  for position, char in enumerate(code):
    if state == DEFAULT:
      # Waiting for comment start character, beginning of string,
      # or escape character.
      if char == '#':
        state = IN_COMMENT
        comment_start_pos = position
      elif char in ('"', "'"):
        string_char = char
        state = IN_STRING
      elif char == '\\':
        state = ESCAPING_CHAR_OUTSIDE_OF_STRING
    elif state == IN_COMMENT:
      if char == '\n':
        yield Comment(markup_span=Span(comment_start_pos, position), text_span=Span(comment_start_pos+1, position), multiline=False)
        current_comment_text = ''
        state = DEFAULT
      else:
        current_comment_text += char
    elif state == IN_STRING:
      if char == string_char:
        state = DEFAULT
      elif char == '\\':
        state = ESCAPING_CHAR_INSIDE_STRING
    elif state == ESCAPING_CHAR_INSIDE_STRING:
      state = IN_STRING
    elif state == ESCAPING_CHAR_OUTSIDE_OF_STRING:
      # Escaping current char, outside of string.
      state = DEFAULT
    if char == '\n':
      line_counter += 1

  # end of file

  if state == IN_COMMENT:
    yield Comment(markup_span=Span(comment_start_pos, position+1), text_span=Span(comment_start_pos+1, position+1), multiline=False)


# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

import unittest
from typing import List

from .. import iter_comments_go, Comment, UnterminatedCommentError


def commentsToList(code:str) -> List[Comment]:
  return list(iter_comments_go(code))


class GoParserTest(unittest.TestCase):

  def testSingleLineComment(self):
    code = '// single line comment'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "// single line comment")
    self.assertEqual(comments[0].text_span.extract(code), " single line comment")
    self.assertEqual(comments[0].multiline, False)

  def testSingleLineCommentInRuneLiteral(self):
    code = "msg := '// this is not a comment'"
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testSingleLineCommentInBackTickedLiteral(self):
    code = "msg := `// this is not a comment`"
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testSingleLineCommentInQuotedLiteral(self):
    code = 'msg := "// this is not a comment"'
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testMultiLineComment(self):
    code = '/* multiline\ncomment */'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '/* multiline\ncomment */')
    self.assertEqual(comments[0].text_span.extract(code), " multiline\ncomment ")
    self.assertEqual(comments[0].multiline, True)

  def testMultiLineCommentWithStars(self):
    code = "/***************/"
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "/***************/")
    self.assertEqual(comments[0].text_span.extract(code), "*************")
    self.assertEqual(comments[0].multiline, True)

  def testMultiLineCommentInRuneLiteral(self):
    code = "msg := '/* This is not a\\nmultiline comment */'"
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testMultiLineCommentInQuotedLiteral(self):
    code = 'msg := "/* This is not a\\nmultiline comment */"'
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testMultiLineCommentInBackTickedLiteral(self):
    code = 'msg := `/* This is not a\\nmultiline comment */`'
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testMultiLineCommentUnterminated(self):
    code = 'a := 1 /* Unterminated\\n comment'
    with self.assertRaises(UnterminatedCommentError):
      commentsToList(code)



from pathlib import Path

from setuptools import setup, find_packages

setup(
  name="commie",
  version="0.0.2",

  author="Art Galkin",
  author_email="ortemeo@gmail.com",
  url='https://github.com/rtmigo/commie.python',

  packages=find_packages(),
  install_requires=[],

  description="Finds comments in source code in different programming languages",

  long_description=(Path(__file__).parent / 'README.md').read_text(),
  long_description_content_type='text/markdown',

  license='MIT',

  # entry_points={
  #      'console_scripts': [
  #           'allrights = allrights:main',
  #       ]},

  keywords=['comments', 'source code'],

  # https://pypi.org/classifiers/
  classifiers=[
    "Development Status :: 2 - Pre-Alpha",
    "Intended Audience :: Developers",
    'License :: OSI Approved :: MIT License',
    'Topic :: Software Development :: Documentation',
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
  ],

  test_suite='nose.collector',
  tests_require=['nose'],
  zip_safe=False
)


# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

"""This module provides methods for parsing comments from C family languages.

Works with:
  C99+
  C++
  Objective-C
  Java

"""

import re
from typing import Iterable

from commie.parsers import common
from commie.parsers.common import Comment
from commie.parsers.helper import matchGroupToComment


def extract_comments(code:str) -> Iterable[Comment]:
  """Extracts a list of comments from the given C family source code.

  Comments are represented with the Comment class found in the common module.
  C family comments come in two forms, single and multi-line comments.
    - Single-line comments begin with '//' and continue to the end of line.
    - Multi-line comments begin with '/*' and end with '*/' and can span
      multiple lines of code. If a multi-line comment does not terminate
      before EOF is reached, then an exception is raised.

  Note that this doesn't take language-specific preprocessor directives into
  consideration.

  Args:
    code: String containing code to extract comments from.
  Returns:
    Python list of common.Comment in the order that they appear in the code.
  Raises:
    common.UnterminatedCommentError: Encountered an unterminated multi-line
      comment.
  """
  pattern = r"""
    (?P<literal> (\"([^\"\n])*\")+) |
    (?P<single> //(?P<single_content>.*)?$) |
    (?P<multi> /\*(?P<multi_content>(.|\n)*?)?\*/) |
    (?P<error> /\*(.*)?)
  """

  compiled = re.compile(pattern, re.VERBOSE | re.MULTILINE)

  for match in compiled.finditer(code):

    kind = match.lastgroup
    #markupSpan = match.span(0)

    if kind == "single":
      yield matchGroupToComment(match, "single_content", False)

    elif kind == "multi":
      yield matchGroupToComment(match, "multi_content", True)

    elif kind == "error":
      raise common.UnterminatedCommentError()



# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

import unittest
from typing import List

from .. import iter_comments_python, Comment


def commentsToList(code:str) -> List[Comment]:
  return list(iter_comments_python(code))


class PythonParserTest(unittest.TestCase):

  def testComment(self):
    code = '# comment'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), '# comment')
    self.assertEqual(comments[0].text_span.extract(code), " comment")
    self.assertEqual(comments[0].multiline, False)

  def testCommentInSingleQuotedString(self):
    code = "'this is # not a comment'"
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testCommentInDoubleQuotedString(self):
    code = '"this is # not a comment"'
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testNestedStringSingleOutside(self):
    code = "'this is \"# not a comment\"'"
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testNestedStringDoubleOutside(self):
    code = '"this is \'# not a comment\'"'
    comments = commentsToList(code)
    self.assertEqual(comments, [])

  def testEscapedSingleQuote(self):
    code = "\\'# this is a comment"
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "# this is a comment")
    self.assertEqual(comments[0].text_span.extract(code), " this is a comment")
    self.assertEqual(comments[0].multiline, False)

  def testEscapedDoubleQuote(self):
    code = '\\"# this is another comment'
    comments = commentsToList(code)

    self.assertEqual(len(comments), 1)

    self.assertEqual(comments[0].markup_span.extract(code), "# this is another comment")
    self.assertEqual(comments[0].text_span.extract(code), " this is another comment")
    self.assertEqual(comments[0].multiline, False)



# SPDX-FileCopyrightText: Copyright (c) 2021 Art Galkin <ortemeo@gmail.com>
# SPDX-FileCopyrightText: Copyright (c) 2015 Jean-Ralph Aviles
# SPDX-License-Identifier: MIT

from typing import Iterable, Optional

from commie.parsers import common
from commie.parsers.common import Comment, Span


def extract_comments(code: str) -> Iterable[Comment]:
  """Extracts a list of comments from the given Javascript source code.

  Comments are represented with the Comment class found in the common module.
  Javascript comments come in two forms, single and multi-line comments.
    - Single-line comments begin with '//' and continue to the end of line.
    - Multi-line comments begin with '/*' and end with '*/' and can span
      multiple lines of code. If a multi-line comment does not terminate
      before EOF is reached, then an exception is raised.
  This module takes quoted strings into account when extracting comments from
  source code.

  Args:
    code: String containing code to extract comments from.
  Returns:
    Python list of common.Comment in the order that they appear in the code.
  Raises:
    common.UnterminatedCommentError: Encountered an unterminated multi-line
      comment.
  """

  WAITING_FOR_COMMENT = 0
  IN_SINGLE_LINE_COMMENT = 2
  IN_MULTI_LINE_COMMENT = 3
  IN_MULTI_LINE_COMMENT_ASTERISK = 4
  IN_STRING = 5
  IN_STRING_ESCAPE_NEXT_CHAR = 6
  FOUND_SLASH = 1

  state = WAITING_FOR_COMMENT
  #current_comment_text = ''

  markup_start_pos = None
  text_start_pos:Optional[int] = 0
  text_length:Optional[int] = 0

  quote = None
  position = 0

  for position, char in enumerate(code):

    if state == WAITING_FOR_COMMENT:
      # Waiting for comment start character or beginning of
      # string.
      if char == '/':
        state = FOUND_SLASH
      elif char in ('"', "'"):
        quote = char
        state = IN_STRING
    elif state == FOUND_SLASH:
      # Found comment start character, classify next character and
      # determine if single or multi-line comment.
      if char == '/':
        state = IN_SINGLE_LINE_COMMENT
        markup_start_pos = position - 1  # we are at second char of "//"
        text_start_pos=position+1
      elif char == '*':
        state = IN_MULTI_LINE_COMMENT
        markup_start_pos = position - 1  # we are at second char of "/*"
        text_start_pos=position+1
      else:
        state = WAITING_FOR_COMMENT
    elif state == IN_SINGLE_LINE_COMMENT:
      # In single-line comment, read characters until EOL.
      if char == '\n':
        yield common.Comment(
          text_span=Span(text_start_pos, text_start_pos+text_length),
          markup_span=Span(markup_start_pos, position+1),
          multiline=False)
        text_length=0
        state = WAITING_FOR_COMMENT
      else:
        text_length+=1
    elif state == IN_MULTI_LINE_COMMENT:
      # In multi-line comment, add characters until '*' is
      # encountered.
      if char == '*':
        state = IN_MULTI_LINE_COMMENT_ASTERISK
      else:
        text_length+=1
    elif state == IN_MULTI_LINE_COMMENT_ASTERISK:
      # In multi-line comment with asterisk found. Determine if
      # comment is ending.
      if char == '/':
        yield Comment(
          text_span=Span(text_start_pos, text_start_pos+text_length),
          markup_span=Span(markup_start_pos, position+1),
          multiline=True)
        text_length=0
        state = WAITING_FOR_COMMENT
      else:
        text_length+=1
        # Care for multiple '*' in a row
        if char != '*':
          text_length+=1
          state = IN_MULTI_LINE_COMMENT
    elif state == IN_STRING:
      # In string literal, expect literal end or escape character.
      if char == quote:
        state = WAITING_FOR_COMMENT
      elif char == '\\':
        state = IN_STRING_ESCAPE_NEXT_CHAR
    elif state == IN_STRING_ESCAPE_NEXT_CHAR:
      state = IN_STRING
    if char == '\n':
      pass

  # end of file
  if state in (IN_MULTI_LINE_COMMENT, IN_MULTI_LINE_COMMENT_ASTERISK):
    raise common.UnterminatedCommentError()

  if state == IN_SINGLE_LINE_COMMENT:
    yield common.Comment(
      text_span=Span(text_start_pos, text_start_pos+text_length),
      markup_span=Span(markup_start_pos, position+1),
      multiline=False)

